create table blockTransaction (id identity not null, txid binary(32) not null, hash binary(32) not null, height integer not null, date timestamp, fee bigint, label varchar(255), transaction varbinary(1048576), blockHash binary(32), wallet bigint not null);
create table blockTransactionHashIndex (id identity not null, hash binary(32) not null, height integer not null, date timestamp, fee bigint, label varchar(255), index bigint not null, outputValue bigint not null, status integer, spentBy bigint, node bigint not null);
create table keystore (id identity not null, label varchar(255), source integer not null, walletModel integer not null, masterFingerprint varchar(8), derivationPath varchar(255) not null, extendedPublicKey varchar(255), masterPrivateExtendedKey bigint, seed bigint, wallet bigint not null, index integer not null);
create table masterPrivateExtendedKey (id identity not null, privateKey varbinary(255), chainCode varbinary(255), initialisationVector varbinary(32), encryptedBytes varbinary(255), keySalt varbinary(32), deriver integer, crypter integer, creationTimeSeconds bigint);
create table policy (id identity not null, name varchar(255) not null, script varchar(2048) not null);
create table seed (id identity not null, type integer not null, mnemonicString varchar(255), initialisationVector varbinary(32), encryptedBytes varbinary(255), keySalt varbinary(32), deriver integer, crypter integer, needsPassphrase boolean, creationTimeSeconds bigint);
create table wallet (id identity not null, name varchar(255) not null, network integer not null, policyType integer not null, scriptType integer not null, storedBlockHeight integer, gapLimit integer, birthDate timestamp, defaultPolicy bigint not null);
create table walletNode (id identity not null, derivationPath varchar(255) not null, label varchar(255), wallet bigint not null, parent bigint);
alter table blockTransactionHashIndex add constraint blockTransactionHashIndex_spentBy_unique unique (spentBy);
alter table keystore add constraint keystore_masterPrivateExtendedKey_unique unique (masterPrivateExtendedKey);
alter table keystore add constraint keystore_seed_unique unique (seed);
alter table wallet add constraint wallet_defaultPolicy_unique unique (defaultPolicy);
alter table blockTransaction add constraint blockTransaction_wallet foreign key (wallet) references wallet;
alter table blockTransactionHashIndex add constraint blockTransactionHashIndex_spentBy foreign key (spentBy) references blockTransactionHashIndex;
alter table blockTransactionHashIndex add constraint blockTransactionHashIndex_node foreign key (node) references walletNode;
alter table keystore add constraint keystore_masterPrivateExtendedKey foreign key (masterPrivateExtendedKey) references masterPrivateExtendedKey;
alter table keystore add constraint keystore_seed foreign key (seed) references seed;
alter table keystore add constraint keystore_wallet foreign key (wallet) references wallet;
alter table wallet add constraint wallet_defaultPolicy foreign key (defaultPolicy) references policy;
alter table walletNode add constraint walletNode_wallet foreign key (wallet) references wallet;
alter table walletNode add constraint walletNode_walletNode foreign key (parent) references walletNode;
create index blockTransaction_txid on blockTransaction(txid);